@@ -8,8 +8,10 @@ from sql import Literal, Null
 from sql.aggregate import Count
 from sql.operators import Concat
 
+from trytond.i18n import gettext
 from trytond.model import Workflow, Model, ModelView, ModelSQL, fields, \
     sequence_ordered
+from trytond.model.exceptions import AccessError
 from trytond.modules.company import CompanyReport
 from trytond.wizard import Wizard, StateAction, StateView, StateTransition, \
     Button
@@ -19,8 +21,11 @@ from trytond.transaction import Transaction
 from trytond.pool import Pool
 
 from trytond.modules.account.tax import TaxableMixin
+from trytond.modules.account_product.exceptions import AccountError
 from trytond.modules.product import price_digits
 
+from .exceptions import PurchaseQuotationError, PartyLocationError
+
 __all__ = ['Purchase', 'PurchaseIgnoredInvoice',
     'PurchaseRecreadtedInvoice', 'PurchaseLine', 'PurchaseLineTax',
     'PurchaseLineIgnoredMove', 'PurchaseLineRecreatedMove', 'PurchaseReport',
@@ -161,8 +166,9 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
             ('paid', 'Paid'),
             ('exception', 'Exception'),
             ], 'Invoice State', readonly=True, required=True)
-    invoices = fields.Function(fields.One2Many('account.invoice', None,
-            'Invoices'), 'get_invoices', searcher='search_invoices')
+    invoices = fields.Function(fields.Many2Many(
+            'account.invoice', None, None, "Invoices"),
+        'get_invoices', searcher='search_invoices')
     invoices_ignored = fields.Many2Many(
             'purchase.purchase-ignored-account.invoice',
             'purchase', 'invoice', 'Ignored Invoices', readonly=True)
@@ -183,27 +189,21 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
             ('received', 'Received'),
             ('exception', 'Exception'),
             ], 'Shipment State', readonly=True, required=True)
-    shipments = fields.Function(fields.One2Many('stock.shipment.in', None,
-            'Shipments'), 'get_shipments', searcher='search_shipments')
-    shipment_returns = fields.Function(
-        fields.One2Many('stock.shipment.in.return', None, 'Shipment Returns'),
+    shipments = fields.Function(fields.Many2Many(
+            'stock.shipment.in', None, None, "Shipments"),
+        'get_shipments', searcher='search_shipments')
+    shipment_returns = fields.Function(fields.Many2Many(
+            'stock.shipment.in.return', None, None, "Shipment Returns"),
         'get_shipment_returns', searcher='search_shipment_returns')
     moves = fields.One2Many('stock.move', 'purchase', 'Moves', readonly=True)
 
     @classmethod
     def __setup__(cls):
         super(Purchase, cls).__setup__()
-        cls.create_date.select = True
         cls._order = [
             ('purchase_date', 'DESC'),
             ('id', 'DESC'),
             ]
-        cls._error_messages.update({
-                'warehouse_required': ('A warehouse must be defined for '
-                    'quotation of purchase "%s".'),
-                'delete_cancel': ('Purchase "%s" must be cancelled before '
-                    'deletion.'),
-                })
         cls._transitions |= set((
                 ('draft', 'quotation'),
                 ('quotation', 'confirmed'),
@@ -247,7 +247,10 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
                     'depends': ['state'],
                     },
                 'process': {
-                    'invisible': Eval('state') != 'confirmed',
+                    'invisible': ~Eval('state').in_(
+                        ['confirmed', 'processing']),
+                    'icon': If(Eval('state') == 'confirmed',
+                        'tryton-forward', 'tryton-refresh'),
                     'depends': ['state'],
                     },
                 'handle_invoice_exception': {
@@ -317,6 +320,9 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
                     values=['processing'],
                     where=sql_table.id.in_(sub_query.select(sub_query.id))))
 
+        # Add index on create_date
+        table.index_action('create_date', action='add')
+
     @classmethod
     def default_payment_term(cls):
         PaymentTerm = Pool().get('account.invoice.payment_term')
@@ -643,7 +649,9 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
             if (not line.to_location
                     and line.product
                     and line.product.type in ('goods', 'assets')):
-                self.raise_user_error('warehouse_required', (self.rec_name,))
+                raise PurchaseQuotationError(
+                    gettext('purchase.msg_warehouse_required_for_quotation',
+                        purchase=self.rec_name))
 
     @classmethod
     def set_number(cls, purchases):
@@ -783,7 +791,9 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
         cls.cancel(purchases)
         for purchase in purchases:
             if purchase.state != 'cancel':
-                cls.raise_user_error('delete_cancel', purchase.rec_name)
+                raise AccessError(
+                    gettext('purchase.msg_purchase_delete_cancel',
+                        purchase=purchase.rec_name))
         super(Purchase, cls).delete(purchases)
 
     @classmethod
@@ -844,8 +854,10 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
     @ModelView.button
     def process(cls, purchases):
         process, done = [], []
-        cls.__lock(purchases)
+        cls.lock(purchases)
         for purchase in purchases:
+            if purchase.state not in {'confirmed', 'processing', 'done'}:
+                continue
             purchase.create_invoice()
             purchase.set_invoice_state()
             purchase.create_move('in')
@@ -865,25 +877,6 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
         if done:
             cls.do(done)
 
-    @classmethod
-    def __lock(cls, records):
-        from trytond.tools import grouped_slice, reduce_ids
-        from sql import Literal, For
-        transaction = Transaction()
-        database = transaction.database
-        connection = transaction.connection
-        table = cls.__table__()
-
-        if database.has_select_for():
-            for sub_records in grouped_slice(records):
-                where = reduce_ids(table.id, sub_records)
-                query = table.select(
-                    Literal(1), where=where, for_=For('UPDATE', nowait=True))
-                with connection.cursor() as cursor:
-                    cursor.execute(*query)
-        else:
-            database.lock(connection, cls._table)
-
     @classmethod
     @ModelView.button_action('purchase.wizard_modify_header')
     def modify_header(cls, purchases):
@@ -956,10 +949,9 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
     product = fields.Many2One('product.product', 'Product',
         ondelete='RESTRICT',
         domain=[
-            ('purchasable', '=', True),
-            If(Bool(Eval('product_supplier')),
-                ('product_suppliers', '=', Eval('product_supplier')),
-                ()),
+            If(Eval('purchase_state').in_(['draft', 'quotation']),
+                [('purchasable', '=', True)],
+                []),
             ],
         states={
             'invisible': Eval('type') != 'line',
@@ -987,8 +979,14 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
         ondelete='RESTRICT',
         domain=[
             If(Bool(Eval('product')),
-                ('product.products', '=', Eval('product')),
-                ()),
+                ['OR',
+                    [
+                        ('template.products', '=', Eval('product')),
+                        ('product', '=', None),
+                        ],
+                    ('product', '=', Eval('product')),
+                    ],
+                []),
             ('party', '=', Eval('_parent_purchase', {}).get('party')),
             ],
         states={
@@ -1048,11 +1046,9 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
     delivery_date = fields.Function(fields.Date('Delivery Date',
             states={
                 'invisible': ((Eval('type') != 'line')
-                    | (Eval('delivery_date_edit', False)
-                        & ~Eval('purchase_state').in_(
-                            ['processing', 'done', 'cancelled']))),
+                    | Eval('delivery_date_edit', False)),
                 },
-            depends=['type', 'delivery_date_edit', 'purchase_state']),
+            depends=['type', 'delivery_date_edit']),
         'on_change_with_delivery_date')
     delivery_date_edit = fields.Boolean(
         "Edit Delivery Date",
@@ -1067,9 +1063,7 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
         "Delivery Date",
         states={
             'invisible': ((Eval('type') != 'line')
-                | ~Eval('delivery_date_edit', False)
-                | Eval('purchase_state').in_(
-                    ['processing', 'done', 'cancelled'])),
+                | ~Eval('delivery_date_edit', False)),
             'readonly': Eval('purchase_state').in_([
                     'processing', 'done', 'cancel']),
             },
@@ -1078,20 +1072,6 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
         fields.Selection(STATES, 'Purchase State'),
         'on_change_with_purchase_state')
 
-    @classmethod
-    def __setup__(cls):
-        super(PurchaseLine, cls).__setup__()
-        cls._error_messages.update({
-                'supplier_location_required': ('Purchase "%(purchase)s" '
-                    'misses the supplier location for line "%(line)s".'),
-                'missing_account_expense': ('Product "%(product)s" of '
-                    'purchase %(purchase)s misses an expense account.'),
-                'missing_default_account_expense': ('Purchase "%(purchase)s" '
-                    'misses a default "account expense".'),
-                'delete_cancel_draft': ('The line "%(line)s" must be on '
-                    'canceled or draft purchase to be deleted.'),
-                })
-
     @classmethod
     def __register__(cls, module_name):
         super(PurchaseLine, cls).__register__(module_name)
@@ -1172,9 +1152,16 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
         context['taxes'] = [t.id for t in self.taxes or []]
         return context
 
+    @fields.depends('purchase', '_parent_purchase.party')
+    def _get_product_supplier_pattern(self):
+        return {
+            'party': self.purchase.party.id if self.purchase.party else -1,
+            }
+
     @fields.depends('product', 'unit', 'quantity', 'purchase',
         '_parent_purchase.party', 'product_supplier',
-        methods=['_get_tax_rule_pattern', '_get_context_purchase_price'])
+        methods=['_get_tax_rule_pattern', '_get_context_purchase_price',
+            '_get_product_supplier_pattern'])
     def on_change_product(self):
         pool = Pool()
         Product = pool.get('product.product')
@@ -1208,14 +1195,12 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
             self.unit = self.product.purchase_uom
             self.unit_digits = self.product.purchase_uom.digits
 
-        if self.purchase and self.purchase.party:
-            product_suppliers = [ps for ps in self.product.product_suppliers
-                if ps.party == self.purchase.party]
-            if len(product_suppliers) == 1:
-                self.product_supplier, = product_suppliers
-        if (self.product_supplier
-                and (self.product_supplier
-                    not in self.product.product_suppliers)):
+        product_suppliers = list(self.product.product_suppliers_used(
+                **self._get_product_supplier_pattern()))
+        if len(product_suppliers) == 1:
+            self.product_supplier, = product_suppliers
+        elif (self.product_supplier
+                and self.product_supplier not in product_suppliers):
             self.product_supplier = None
 
         with Transaction().set_context(self._get_context_purchase_price()):
@@ -1231,9 +1216,12 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
     @fields.depends('product', 'product_supplier',
         methods=['on_change_product'])
     def on_change_product_supplier(self):
-        if not self.product and self.product_supplier:
-            if len(self.product_supplier.product.products) == 1:
-                self.product, = self.product_supplier.product.products
+        if self.product_supplier:
+            if self.product_supplier.product:
+                self.product = self.product_supplier.product
+            elif not self.product:
+                if len(self.product_supplier.template.products) == 1:
+                    self.product, = self.product_supplier.template.products
         self.on_change_product()
 
     @fields.depends('product')
@@ -1294,8 +1282,7 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
 
     def get_from_location(self, name):
         if (self.quantity or 0) >= 0:
-            if self.purchase.party.supplier_location:
-                return self.purchase.party.supplier_location.id
+            return self.purchase.party.supplier_location.id
         elif self.purchase.return_from_location:
             return self.purchase.return_from_location.id
 
@@ -1303,7 +1290,7 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
         if (self.quantity or 0) >= 0:
             if self.purchase.warehouse:
                 return self.purchase.warehouse.input_location.id
-        elif self.purchase.party.supplier_location:
+        else:
             return self.purchase.party.supplier_location.id
 
     @fields.depends('product_supplier', 'quantity', 'moves', 'purchase',
@@ -1313,10 +1300,10 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
     def on_change_with_delivery_date(self, name=None):
         pool = Pool()
         Date = pool.get('ir.date')
-        moves = [m for m in self.moves if m.state != 'cancel']
-        if moves:
+        if self.moves:
             dates = filter(
-                None, (m.effective_date or m.planned_date for m in moves))
+                None, (m.effective_date or m.planned_date for m in self.moves
+                    if m.state != 'cancel'))
             return min(dates, default=None)
         delivery_date = None
         if self.delivery_date_edit:
@@ -1350,8 +1337,6 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
 
         invoice_line = InvoiceLine()
         invoice_line.type = self.type
-        invoice_line.currency = self.purchase.currency
-        invoice_line.company = self.purchase.company
         invoice_line.description = self.description
         invoice_line.note = self.note
         invoice_line.origin = self
@@ -1379,16 +1364,19 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
         if self.product:
             invoice_line.account = self.product.account_expense_used
             if not invoice_line.account:
-                self.raise_user_error('missing_account_expense', {
-                        'product': invoice_line.product.rec_name,
-                        'purchase': self.purchase.rec_name,
-                        })
+                raise AccountError(
+                    gettext('purchase'
+                        '.msg_purchase_product_missing_account_expense',
+                        purchase=self.purchase.rec_name,
+                        product=self.product.rec_name))
         else:
             invoice_line.account = account_config.get_multivalue(
                 'default_category_account_expense')
             if not invoice_line.account:
-                self.raise_user_error('missing_default_account_expense',
-                    {'purchase': self.purchase.rec_name})
+                raise AccountError(
+                    gettext('purchase'
+                        '.msg_purchase_missing_account_expense',
+                        purchase=self.purchase.rec_name))
         invoice_line.stock_moves = self._get_invoice_line_moves()
         return [invoice_line]
 
@@ -1468,10 +1456,10 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
             return
 
         if not self.purchase.party.supplier_location:
-            self.raise_user_error('supplier_location_required', {
-                    'purchase': self.purchase.rec_name,
-                    'line': self.rec_name,
-                    })
+            raise PartyLocationError(
+                gettext('purchase.msg_purchase_supplier_location_required',
+                    purchase=self.purchase.rec_name,
+                    party=self.purchase.party.rec_name))
         move = Move()
         move.quantity = quantity
         move.uom = self.unit
@@ -1542,9 +1530,10 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
     def delete(cls, lines):
         for line in lines:
             if line.purchase_state not in {'cancel', 'draft'}:
-                cls.raise_user_error('delete_cancel_draft', {
-                        'line': line.rec_name,
-                        })
+                raise AccessError(
+                    gettext('purchase.msg_purchase_line_delete_cancel_draft',
+                        line=line.rec_name,
+                        purchase=line.purchase.rec_name))
         super(PurchaseLine, cls).delete(lines)
 
     @classmethod
@@ -1780,24 +1769,15 @@ class ModifyHeader(Wizard):
             ])
     modify = StateTransition()
 
-    @classmethod
-    def __setup__(cls):
-        super(ModifyHeader, cls).__setup__()
-        cls._error_messages.update({
-                'not_in_draft': (
-                    'The purchase "%(purchase)s" must be in draft '
-                    'to modify header.'),
-                })
-
     def get_purchase(self):
         pool = Pool()
         Purchase = pool.get('purchase.purchase')
 
         purchase = Purchase(Transaction().context['active_id'])
         if purchase.state != 'draft':
-            self.raise_user_error('not_in_draft', {
-                    'purchase': purchase.rec_name,
-                    })
+            raise AccessError(
+                gettext('purchase.msg_purchase_modify_header_draft',
+                    purchase=purchase.rec_name))
         return purchase
 
     def default_start(self, fields):
